# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Denis VIVIÈS <dvivies@geoblink.com>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.

from logging import getLogger

import pytest
import sqlalchemy
from sqlalchemy import event

from anyblok.config import Configuration
from anyblok.registry import RegistryManager
from anyblok.blok import BlokManager
from anyblok.testing import load_configuration

logger = getLogger(__name__)


@pytest.fixture(scope='session')
def configuration_loaded(request):
    load_configuration()


@pytest.fixture(scope='session')
def init_session(request, configuration_loaded):
    # Init registry
    additional_setting = {'unittest': True}

    if len(BlokManager.list()) == 0:
        BlokManager.load()
    registry = RegistryManager.get(Configuration.get('db_name'),
                                   **additional_setting)
    registry.commit()
    # Add savepoint
    registry.begin_nested()

    @event.listens_for(registry.session, 'after_transaction_end')
    def restart_savepoint(session, transaction):
        if transaction.nested and not transaction._parent.nested:
            session.expire_all()
            session.begin_nested()

    logger.info('Creating registry')
    yield registry

    request.addfinalizer(registry.session.close)


@pytest.fixture(scope='function')
def rollback_registry(request, init_session):
    registry = init_session

    yield registry

    def clean_up():
        try:
            logger.debug('Invalidating all cache')
            registry.System.Cache.invalidate_all()
        except sqlalchemy.exc.InvalidRequestError:
            logger.warning('Invalid request Error: while invalidating all '
                           'caches after {}'
                           .format(request.function.__name__))
        finally:
            registry.rollback()

    request.addfinalizer(clean_up)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.version import parse_version


class TestParseVersion:

    def test_gt(self):
        version = parse_version('1.2.3')
        assert version > '1.1.13'
        assert not (version > '1.12.3')
        assert not (version > '1.2.3')

    def test_ge(self):
        version = parse_version('1.2.3')
        assert version >= '1.1.13'
        assert not (version >= '1.12.3')
        assert version >= '1.2.3'

    def test_eq(self):
        version = parse_version('1.2.3')
        assert not (version == '1.1.13')
        assert not (version == '1.12.3')
        assert version == '1.2.3'

    def test_ne(self):
        version = parse_version('1.2.3')
        assert version != '1.1.13'
        assert version != '1.12.3'
        assert not (version != '1.2.3')

    def test_lt(self):
        version = parse_version('1.2.3')
        assert not (version < '1.1.13')
        assert version < '1.12.3'
        assert not (version < '1.2.3')

    def test_le(self):
        version = parse_version('1.2.3')
        assert not (version <= '1.1.13')
        assert version <= '1.12.3'
        assert version <= '1.2.3'

    def test_wrong_version(self):
        parse_version('Wrong version')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from sqlalchemy.schema import ForeignKey
from anyblok.mapper import (ModelAttribute, ModelAttributeException,
                            ModelAttributeAdapter, ModelAdapter, ModelRepr,
                            ModelReprException, ModelAttributeAdapterException,
                            ModelMapper, ModelAttributeMapper, MapperAdapter,
                            MapperException)
from anyblok import Declarations
from anyblok.column import String, Integer
from .conftest import init_registry, reset_db

register = Declarations.register
unregister = Declarations.unregister
Model = Declarations.Model


class TestModelAttribute:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_blok):
        transaction = registry_blok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_get_attribute(self, registry_blok):
        registry = registry_blok
        ma = ModelAttribute('Model.System.Model', 'name')
        assert ma.get_attribute(registry) == registry.get(
            'Model.System.Model').name

    def test_without_attribute(self, registry_blok):
        with pytest.raises(ModelAttributeException):
            ModelAttribute('Model.System.Model', None)

    def test_without_model(self, registry_blok):
        with pytest.raises(ModelAttributeException):
            ModelAttribute(None, 'name')

    def test_without_model_and_attribute(self, registry_blok):
        with pytest.raises(ModelAttributeException):
            ModelAttribute(None, None)

    def test_get_attribute_unexisting_model(self, registry_blok):
        registry = registry_blok
        ma = ModelAttribute('Model.Unexisting.Model', 'name')
        with pytest.raises(ModelAttributeException):
            ma.get_attribute(registry)

    def test_get_attribute_unexisting_attribute(self, registry_blok):
        registry = registry_blok
        ma = ModelAttribute('Model.System.Model', 'id')
        with pytest.raises(ModelAttributeException):
            ma.get_attribute(registry)

    def test_get_fk_name_with_wrong_model(self, registry_blok):
        registry = registry_blok
        ma = ModelAttribute('Model.System', 'name')
        with pytest.raises(ModelAttributeException):
            ma.get_fk_name(registry)

    def test_get_fk_name_with_unexisting_model(self, registry_blok):
        registry = registry_blok
        ma = ModelAttribute('Model.Unexisting.Model', 'name')
        with pytest.raises(ModelAttributeException):
            ma.get_fk_name(registry)

    def test_get_fk_name_unexisting_attribute(self, registry_blok):
        registry = registry_blok
        ma = ModelAttribute('Model.System.Model', 'id')
        with pytest.raises(ModelAttributeException):
            ma.get_fk_name(registry)

    def test_get_fk_name(self, registry_blok):
        registry = registry_blok
        ma = ModelAttribute('Model.System.Model', 'name')
        assert ma.get_fk_name(registry) == 'system_model.name'

    def test_get_fk(self, registry_blok):
        registry = registry_blok
        ma = ModelAttribute('Model.System.Model', 'name')
        assert isinstance(ma.get_fk(registry), ForeignKey)

    def test_get_fk_with_options(self, registry_blok):
        registry = registry_blok
        ma = ModelAttribute('Model.System.Model', 'name').options(
            ondelete='cascade')
        mafk = ma.get_fk(registry)
        assert isinstance(mafk, ForeignKey)

    def test_use(self, registry_blok):
        ma = Declarations.Model.System.Model.use('name')
        assert isinstance(ma, ModelAttribute)
        assert ma.model_name == 'Model.System.Model'
        assert ma.attribute_name == 'name'


class TestModelAttributeAdapter:

    def test_from_declaration(self):
        ma = ModelAttribute('Model.System.Model', 'name')
        maa = ModelAttributeAdapter(ma)
        assert maa is ma

    def test_from_registry_name(self):
        maa = ModelAttributeAdapter("Model.System.Model=>name")
        assert isinstance(maa, ModelAttribute)
        assert maa.model_name == 'Model.System.Model'
        assert maa.attribute_name == 'name'

    def test_from_registry_name_without_attribute(self):
        with pytest.raises(ModelAttributeAdapterException):
            ModelAttributeAdapter("Model.System.Model")


class TestModelRepr:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_blok):
        transaction = registry_blok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_unexisting_model(self, registry_blok):
        registry = registry_blok
        mr = ModelRepr('Model.Unexisting.Model')
        with pytest.raises(ModelReprException):
            mr.check_model(registry)

    def test_get_tablename(self, registry_blok):
        registry = registry_blok
        mr = ModelRepr('Model.System.Model')
        assert mr.tablename(registry) == 'system_model'

    def test_get_registry_name(self, registry_blok):
        mr = ModelRepr('Model.System.Model')
        assert mr.model_name == 'Model.System.Model'

    def test_get_primary_keys(self, registry_blok):
        registry = registry_blok
        mr = ModelRepr('Model.System.Model')
        mas = mr.primary_keys(registry)
        assert len(mas) == 1
        assert [x.attribute_name for x in mas] == ['name']

    def test_without_model(self, registry_blok):
        with pytest.raises(ModelReprException):
            ModelRepr(None)


class TestModelAdapter:

    def test_from_declaration(self):
        mr = ModelRepr('Model.System.Model')
        mra = ModelAdapter(mr)
        assert mr is mra

    def test_from_registry_name(self):
        mra = ModelAdapter("Model.System.Model")
        assert isinstance(mra, ModelRepr)
        assert mra.model_name == 'Model.System.Model'


class TestModelMapper:

    def test_not_capable(self):
        assert not (ModelMapper.capable(None))

    def test_capable_by_declaration(self):
        assert ModelMapper.capable(Model.System.Model)

    def test_capable_by_registry_name(self):
        assert ModelMapper.capable('Model.System.Model')

    def test_by_declaration(self):
        mm = ModelMapper(Model.System.Model, 'even')
        assert isinstance(mm.model, ModelRepr)
        assert mm.model.model_name, 'Model.System.Model'

    def test_by_registry_name(self):
        mm = ModelMapper('Model.System.Model', 'event')
        assert isinstance(mm.model, ModelRepr)
        assert mm.model.model_name, 'Model.System.Model'

    def test_listen_sqlalchemy(self):

        def method():
            pass

        mm = ModelMapper(Model.System.Model, 'before_insert')
        mm.listen(method)
        assert method.is_an_sqlalchemy_event_listener
        assert method.sqlalchemy_listener is mm

    def test_listen_anyblok(self):
        def method():
            pass

        mm = ModelMapper(Model.System.Model, 'event')
        mm.listen(method)
        assert method.is_an_event_listener
        assert method.model == 'Model.System.Model'
        assert method.event == 'event'


class TestModelAttributeMapper:

    def test_not_str_capable(self):
        assert not (ModelAttributeMapper.capable('Model.System.Model'))

    def test_not_capable(self):
        assert not (ModelAttributeMapper.capable(None))

    def test_capable_by_declaration(self):
        assert ModelAttributeMapper.capable(
            Model.System.Model.use('name'))

    def test_capable_by_registry_name(self):
        assert ModelAttributeMapper.capable(
            'Model.System.Model=>name')

    def test_by_declaration(self):
        mam = ModelAttributeMapper(Model.System.Model.use('name'), 'event')
        assert isinstance(mam.attribute, ModelAttribute)
        assert mam.attribute.model_name == 'Model.System.Model'

    def test_by_registry_name(self):
        mam = ModelAttributeMapper('Model.System.Model=>name', 'event')
        assert isinstance(mam.attribute, ModelAttribute)
        assert mam.attribute.model_name == 'Model.System.Model'

    def test_listen(self):

        def method():
            pass

        mam = ModelAttributeMapper(Model.System.Model.use('name'), 'set')
        mam.listen(method)
        assert method.is_an_sqlalchemy_event_listener
        assert method.sqlalchemy_listener is mam


class TestMapperAdapter:

    def test_model_mapper(self):
        mam = MapperAdapter('Model.System.Model', 'event')
        assert isinstance(mam, ModelMapper)

    def test_model_attribute_mapper(self):
        mam = MapperAdapter('Model.System.Model=>name', 'event')
        assert isinstance(mam, ModelAttributeMapper)

    def test_no_mapper(self):
        with pytest.raises(MapperException):
            MapperAdapter(None)


class TestMapperOther:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        reset_db()
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_get_fk_name_with_name_different_of_column_name(self):

        def add_in_registry():

            @register(Model)
            class Test:

                name = String(primary_key=True, db_column_name='other')

        registry = self.init_registry(add_in_registry)
        ma = ModelAttribute('Model.Test', 'name')
        assert ma.get_fk_name(registry) == 'test.other'

    def test_get_foreign_key_for(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(foreign_key=Model.Test.use('id'))

        registry = self.init_registry(add_in_registry)
        mr = ModelRepr('Model.Test2')
        mas = mr.foreign_keys_for(registry, 'Model.Test')
        assert len(mas) == 1
        assert [x.attribute_name for x in mas] == ['test_id']

    def test_get_mapper(self):
        registry = self.init_registry(None)
        mm = ModelMapper(Model.System.Model, 'even')
        assert mm.mapper(registry, None) is registry.System.Model

    def test_get_attribute_mapper(self):
        registry = self.init_registry(None)
        mam = ModelAttributeMapper(Model.System.Model.use('name'), 'set')
        # We can't compare that the column are the same, because is the case
        # of the call are in the class attribute (no instance) SQLAlchemy
        # Wrap the result for each call, then each call return a différent
        # object, but it is not an error
        getted = str(mam.mapper(registry, None) == 'test')
        wanted = str(registry.System.Model.name == 'test')
        assert getted == wanted



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .exceptions import ModelFactoryException
from anyblok.field import Field, FieldException
from sqlalchemy.sql import table, and_
from sqlalchemy.orm import Query, mapper, relationship
from .exceptions import ViewException
from anyblok.common import anyblok_column_prefix
from .view import CreateView, DropView


def has_sql_fields(bases):
    """ Tells whether the model as field or not

    :param bases: list of Model's Class
    :rtype: boolean
    """
    for base in bases:
        for p in base.__dict__.keys():
            try:
                if hasattr(getattr(base, p), '__class__'):
                    if Field in getattr(base, p).__class__.__mro__:
                        return True
            except FieldException:
                # field function case already computed
                return True

    return False


class BaseFactory:

    def __init__(self, registry):
        self.registry = registry

    def insert_core_bases(self, bases, properties):
        raise ModelFactoryException('Must be overwritten')

    def build_model(self, modelname, bases, properties):
        raise ModelFactoryException('Must be overwritten')


class ModelFactory(BaseFactory):

    def insert_core_bases(self, bases, properties):
        if has_sql_fields(bases):
            bases.extend(
                [x for x in self.registry.loaded_cores['SqlBase']])
            bases.append(self.registry.declarativebase)
        else:
            # remove tablename to inherit from a sqlmodel
            del properties['__tablename__']

        bases.extend([x for x in self.registry.loaded_cores['Base']])

    def build_model(self, modelname, bases, properties):
        return type(modelname, tuple(bases), properties)


def get_columns(view, columns):
    if not isinstance(columns, list):
        if ', ' in columns:
            columns = columns.split(', ')
        else:
            columns = [columns]

    return [getattr(view.c, x.split(' => ')[1]) for x in columns]


class ViewFactory(BaseFactory):

    def insert_core_bases(self, bases, properties):
        bases.extend(
            [x for x in self.registry.loaded_cores['SqlViewBase']])
        bases.extend([x for x in self.registry.loaded_cores['Base']])

    def build_model(self, modelname, bases, properties):
        Model = type(modelname, tuple(bases), properties)
        self.apply_view(Model, properties)
        return Model

    def apply_view(self, base, properties):
        """ Transform the sqlmodel to view model

        :param base: Model cls
        :param properties: properties of the model
        :exception: MigrationException
        :exception: ViewException
        """
        tablename = base.__tablename__
        if hasattr(base, '__view__'):
            view = base.__view__
        elif tablename in self.registry.loaded_views:
            view = self.registry.loaded_views[tablename]
        else:
            if not hasattr(base, 'sqlalchemy_view_declaration'):
                raise ViewException(
                    "%r.'sqlalchemy_view_declaration' is required to "
                    "define the query to apply of the view" % base)

            view = table(tablename)

            self.registry.loaded_views[tablename] = view
            selectable = getattr(base, 'sqlalchemy_view_declaration')()

            if isinstance(selectable, Query):
                selectable = selectable.subquery()

            for c in selectable.c:
                c._make_proxy(view)

            DropView(tablename).execute_at(
                'before-create', self.registry.declarativebase.metadata)
            CreateView(tablename, selectable).execute_at(
                'after-create', self.registry.declarativebase.metadata)
            DropView(tablename).execute_at(
                'before-drop', self.registry.declarativebase.metadata)

        pks = [col for col in properties['loaded_columns']
               if getattr(getattr(base, anyblok_column_prefix + col),
                          'primary_key', False)]

        if not pks:
            raise ViewException(
                "%r have any primary key defined" % base)

        pks = [getattr(view.c, x) for x in pks]

        mapper_properties = self.get_mapper_properties(base, view, properties)
        setattr(base, '__view__', view)
        __mapper__ = mapper(
            base, view, primary_key=pks, properties=mapper_properties)
        setattr(base, '__mapper__', __mapper__)

    def get_mapper_properties(self, base, view, properties):
        mapper_properties = base.define_mapper_args()
        for field in properties['loaded_columns']:
            if not hasattr(properties[anyblok_column_prefix + field],
                           'anyblok_field'):
                mapper_properties[field] = getattr(view.c, field)
                continue

            anyblok_field = properties[
                anyblok_column_prefix + field].anyblok_field
            kwargs = anyblok_field.kwargs.copy()
            if 'foreign_keys' in kwargs:
                foreign_keys = kwargs['foreign_keys'][1:][:-1].split(', ')
                foreign_keys = [getattr(view.c, x.split('.')[1])
                                for x in foreign_keys]
                kwargs['foreign_keys'] = foreign_keys

            if anyblok_field.model.model_name == base.__registry_name__:
                remote_columns = get_columns(
                    view, kwargs['info']['remote_columns'])
                local_columns = get_columns(
                    view, kwargs['info']['local_columns'])

                assert len(remote_columns) == len(local_columns)
                primaryjoin = []
                for i in range(len(local_columns)):
                    primaryjoin.append(remote_columns[i] == local_columns[i])

                if len(primaryjoin) == 1:
                    primaryjoin = primaryjoin[0]
                else:
                    primaryjoin = and_(*primaryjoin)

                kwargs['remote_side'] = remote_columns
                kwargs['primaryjoin'] = primaryjoin
                Model = base
            else:
                Model = self.registry.get(anyblok_field.model.model_name)

            mapper_properties[field] = relationship(Model, **kwargs)

        return mapper_properties



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Denis VIVIÈS <dvivies@geoblink.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest


@pytest.mark.usefixtures('rollback_registry')
class TestSystemBlok:
    def test_list_by_state_installed(self, rollback_registry):
        registry = rollback_registry
        installed = registry.System.Blok.list_by_state('installed')
        core_is_installed = 'anyblok-core' in installed
        assert core_is_installed is True

    def test_list_by_state_without_state(self, rollback_registry):
        registry = rollback_registry
        assert registry.System.Blok.list_by_state() is None



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.registry import RegistryManager
from anyblok.blok import BlokManager
from anyblok.model import Model
from anyblok.environment import EnvironmentManager


class TestRegistryManager:

    @pytest.fixture(autouse=True)
    def revert_registry_manager(self, request):
        def revert():
            RegistryManager.undeclare_entry('Other')

        request.addfinalizer(revert)

    def test_declared_entries(self):
        hasModel = 'Model' in RegistryManager.declared_entries
        hasMixin = 'Mixin' in RegistryManager.declared_entries
        assert hasModel
        assert hasMixin

    def test_init_blok(self):
        RegistryManager.init_blok('newblok')
        is_exist = 'newblok' in RegistryManager.loaded_bloks
        assert is_exist
        for core in ('Base', 'SqlBase', 'SqlViewBase', 'Session', 'Query',
                     'InstrumentedList'):
            assert core in RegistryManager.loaded_bloks[
                'newblok']['Core'].keys()
            assert RegistryManager.loaded_bloks['newblok']['Core'][core] == []

        assert RegistryManager.loaded_bloks['newblok']['Model'] == {
            'registry_names': []}
        assert RegistryManager.loaded_bloks['newblok']['Mixin'] == {
            'registry_names': []}

    def test_init_blok_with_other_entry(self):
        RegistryManager.declare_entry('Other')
        hasOther = 'Other' in RegistryManager.declared_entries
        assert hasOther
        RegistryManager.init_blok('newblok')
        is_exist = 'newblok' in RegistryManager.loaded_bloks
        assert is_exist
        for core in ('Base', 'SqlBase', 'SqlViewBase', 'Session', 'Query',
                     'InstrumentedList'):
            assert core in RegistryManager.loaded_bloks[
                'newblok']['Core'].keys()
            assert RegistryManager.loaded_bloks['newblok']['Core'][core] == []

        assert RegistryManager.loaded_bloks['newblok']['Model'] == {
            'registry_names': []}
        assert RegistryManager.loaded_bloks['newblok']['Mixin'] == {
            'registry_names': []}
        assert RegistryManager.loaded_bloks['newblok']['Other'] == {
            'registry_names': []}

    def test_anyblok_core_loaded(self):
        BlokManager.load()
        is_exist = 'anyblok-core' in RegistryManager.loaded_bloks
        assert is_exist
        anyblokcore = RegistryManager.loaded_bloks['anyblok-core']
        assert len(anyblokcore['Core']['Base']) == 1
        assert len(anyblokcore['Core']['SqlBase']) == 1
        assert len(anyblokcore['Core']['SqlViewBase']) == 1
        assert len(anyblokcore['Core']['Session']) == 1
        assert len(anyblokcore['Core']['Query']) == 1
        assert len(anyblokcore['Core']['InstrumentedList']) == 1
        is_exist = 'Model.System' in anyblokcore['Model']
        assert is_exist
        BlokManager.unload()

    def test_add_entry(self):
        RegistryManager.declare_entry('Other')
        RegistryManager.init_blok('newblok')
        is_exist = 'newblok' in RegistryManager.loaded_bloks
        assert is_exist
        for entry in ('Base', 'SqlBase', 'SqlViewBase', 'Session', 'Query',
                      'InstrumentedList'):
            assert RegistryManager.loaded_bloks['newblok']['Core'][entry] == []

        assert RegistryManager.loaded_bloks['newblok']['Model'] == {
            'registry_names': []}
        assert RegistryManager.loaded_bloks['newblok']['Mixin'] == {
            'registry_names': []}
        assert RegistryManager.loaded_bloks['newblok']['Other'] == {
            'registry_names': []}

    def test_add_callback(self):

        def callback():
            pass

        RegistryManager.declare_entry('Other', assemble_callback=callback,
                                      initialize_callback=callback)
        hasModel = 'Model' in RegistryManager.declared_entries
        hasMixin = 'Mixin' in RegistryManager.declared_entries
        hasOther = 'Other' in RegistryManager.declared_entries
        assert hasModel
        assert hasMixin
        assert hasOther

        cb = Model.assemble_callback
        hasModelCb = cb == RegistryManager.callback_assemble_entries['Model']
        cb = callback
        hasOtherCb = cb == RegistryManager.callback_assemble_entries['Other']
        assert hasModelCb
        assert hasOtherCb

        cb = Model.initialize_callback
        hasModelCb = cb == RegistryManager.callback_initialize_entries['Model']
        cb = callback
        hasOtherCb = cb == RegistryManager.callback_initialize_entries['Other']
        assert hasModelCb
        assert hasOtherCb

        RegistryManager.init_blok('newblok')
        is_exist = 'newblok' in RegistryManager.loaded_bloks
        assert is_exist
        hasCore = 'Core' in RegistryManager.loaded_bloks['newblok']
        hasModel = 'Model' in RegistryManager.loaded_bloks['newblok']
        hasMixin = 'Mixin' in RegistryManager.loaded_bloks['newblok']
        hasOther = 'Other' in RegistryManager.loaded_bloks['newblok']
        assert hasCore
        assert hasModel
        assert hasMixin
        assert hasOther

    def test_reload_blok(self):
        BlokManager.load()
        try:
            RegistryManager.reload()
        finally:
            BlokManager.unload()

    def test_global_property(self):
        RegistryManager.declare_entry('Other')
        blok = 'newblok'
        RegistryManager.init_blok(blok)
        try:
            oldblok = EnvironmentManager.get('current_blok')
            EnvironmentManager.set('current_blok', blok)
            assert not RegistryManager.has_blok_property('myproperty')
            RegistryManager.add_or_replace_blok_property('myproperty', 2)
            assert RegistryManager.has_blok_property('myproperty')
            assert RegistryManager.get_blok_property('myproperty') == 2
            RegistryManager.add_or_replace_blok_property('myproperty', 3)
            assert RegistryManager.get_blok_property('myproperty') == 3
            RegistryManager.remove_blok_property('myproperty')
            assert not RegistryManager.has_blok_property('myproperty')
        finally:
            EnvironmentManager.set('current_blok', oldblok)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Denis VIVIÈS <dvivies@geoblink.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest

from ..exceptions import CoreBaseException


@pytest.mark.usefixtures('rollback_registry')
class TestCoreBaseScope:
    def test_to_primary_keys(self, rollback_registry):
        registry = rollback_registry
        with pytest.raises(CoreBaseException):
            test = registry.System()
            test.to_primary_keys()

    def test_from_primary_keys(self, rollback_registry):
        registry = rollback_registry
        with pytest.raises(CoreBaseException):
            registry.System.from_primary_keys()

    def test_get_primary_keys(self, rollback_registry):
        registry = rollback_registry
        with pytest.raises(CoreBaseException):
            registry.System.get_primary_keys()

    def test_get_model(self, rollback_registry):
        registry = rollback_registry
        m = registry.System.Model
        m2 = registry.System.get_model('Model.System.Model')
        assert m == m2



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.

from ..testing import (  # noqa
    TestCase, DBTestCase, BlokTestCase, SharedDataTestCase, LogCapture,
    skip_unless_bloks_installed, sgdb_in, load_configuration)


load_configuration()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok


class TestBlok(Blok):

    version = '1.0.0'
    table_exist_before_automigration = None
    table_exist_after_automigration = None

    @classmethod
    def import_declaration_module(cls):
        from . import test  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import test
        reload(test)

    def is_table_exist(self):
        drivername = self.registry.engine.url.drivername
        if drivername.startswith('postgres'):
            return bool(self.registry.execute("""
                select count(*)
                from pg_catalog.pg_tables
                where tablename = 'test';
            """).fetchone()[0])
        elif drivername.startswith('mysql'):
            return bool(self.registry.execute("""
                show tables like 'test';
            """).fetchall())

    def pre_migration(self, latest_version):
        self.__class__.table_exist_before_automigration = self.is_table_exist()

    def post_migration(self, latest_version):
        self.__class__.table_exist_after_automigration = self.is_table_exist()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.

"""This module's main purpose is to test the tests base classes.

Testing mod_authz itself is not the primary objective, we simply needed a
Blok and a very simple, non structural model, to exert the tests base
classes.
"""
import unittest
from anyblok.tests.testcase import SharedDataTestCase, sgdb_in
from anyblok.tests.testcase import BlokTestCase


class Helper:
    """A mixin to be added to testing test classes.

    The purpose of this mixin is to provide a high-level API to create
    and test registry and database conditions.
    """

    @classmethod
    def break_registry(cls):
        """Voluntarily break the database connection."""
        Blok = cls.registry.System.Blok
        # this is a Programming Error (type is wrong)
        # without proper rollbacking, any test running after that
        # would be in error, too.
        Blok.query().filter(Blok.name == 2).all()

    def assert_pristine_registry(self):
        """Assert that the registry works and has no leftover test data."""
        # always True, but needs the DB and registry to work
        self.assertEqual(
            self.registry.System.Blok.query().filter_by(
                name='no-such-blok').count(),
            0)
        # actual rollback of shared test data
        Grant = self.registry.Authorization.ModelPermissionGrant
        self.assertEqual(Grant.query().count(), 0)


class TestSharedDataTestCaseException(unittest.TestCase):
    """Test class with an exception in setUpSharedData.

    We want to ensure that the test is indeed an error, but the
    database connection is not broken.

    In these tests, we use :class:`unittest.TestSuite` to perform the
    various stages of setup (and avoid relying too much on unittest
    internal API). Also :class:`unittest.TestSuite` promises to run the
    tests in the order they were added (we double check that anyway).
    """

    def test_without_error(self):
        """Test normal rollbacks of individual test cases and class."""

        run_order = []

        class SharedData(SharedDataTestCase, Helper):

            @classmethod
            def setUpSharedData(cls):
                cls.Grant = cls.registry.Authorization.ModelPermissionGrant
                cls.grant = cls.Grant.insert(model='abc',
                                             principal='me',
                                             permission='read')

            def test_normal(self):
                """Changes the data, and should be rollbacked in tearDown."""
                run_order.append('normal')
                self.assertEqual(self.grant.principal, 'me')
                self.grant.principal = 'changed'
                self.registry.flush()
                self.assertEqual(
                    self.Grant.query().filter_by(principal='changed').count(),
                    1)

            def test_rollbacked(self):
                """Previous test has been correctly rollbacked."""
                run_order.append('rollbacked')
                self.assertEqual(self.grant.principal, 'me')
                self.assertEqual(
                    self.Grant.query().filter_by(principal='me').count(),
                    1)

        class AfterSharedData(BlokTestCase, Helper):

            def test_class_rollbacked(self):
                run_order.append('class_rollbacked')
                self.assert_pristine_registry()

        result = unittest.TestResult()

        suite = unittest.TestSuite()
        suite.addTest(SharedData('test_normal'))
        suite.addTest(SharedData('test_rollbacked'))
        suite.addTest(AfterSharedData('test_class_rollbacked'))
        suite.run(result)

        self.assertEqual(run_order,
                         ['normal', 'rollbacked', 'class_rollbacked'])
        self.assertEqual(result.testsRun, 3)
        self.assertFalse(result.failures)
        self.assertFalse(result.errors)

    @unittest.skipIf(sgdb_in(['MySQL', 'MariaDB']),
                     "ShareDataTestCase doesn't work, issue #96")
    def test_setup_error(self):
        """Test rollback of test case and class with error in setUp"""

        run_order = []

        class SharedData(SharedDataTestCase, Helper):

            @classmethod
            def setUpSharedData(cls):
                cls.Grant = cls.registry.Authorization.ModelPermissionGrant
                cls.grant = cls.Grant.insert(model='abc',
                                             principal='me',
                                             permission='read')

            def setUp(self):
                run_order.append('error')
                super(SharedData, self).setUp()
                self.break_registry()

            def test_error(self):
                """Won't even be executed (error is in setUp)"""

        class AfterSharedData(BlokTestCase, Helper):

            def test_class_rollbacked(self):
                run_order.append('class_rollbacked')
                self.assert_pristine_registry()

        result = unittest.TestResult()

        suite = unittest.TestSuite()
        suite.addTest(SharedData('test_error'))
        suite.addTest(AfterSharedData('test_class_rollbacked'))
        suite.run(result)

        self.assertEqual(run_order, ['error', 'class_rollbacked'])
        self.assertEqual(result.testsRun, 2)
        self.assertEqual(len(result.errors), 1)
        self.assertFalse(result.failures)

    def test_setup_shared_data_error(self):
        """Test recovery of database errors in setUpSharedData"""

        run_order = []

        class BadTest(SharedDataTestCase):

            @classmethod
            def setUpSharedData(cls):
                run_order.append('test_error')
                cls.break_registry()

            def test_error(cls):
                """Won't even be executed (is after error in setUpClass)"""

        class GoodTest(BlokTestCase, Helper):

            def test_ok(self):
                """Test that registry and db connection work normally."""
                run_order.append('test_ok')
                self.assert_pristine_registry()

        result = unittest.TestResult()

        suite = unittest.TestSuite()
        suite.addTest(BadTest('test_error'))
        suite.addTest(GoodTest('test_ok'))
        suite.run(result)

        self.assertEqual(run_order, ['test_error', 'test_ok'])
        # only test_ok could actually be run
        self.assertEqual(result.testsRun, 1)
        # but it succeeded
        self.assertEqual(len(result.errors), 1)
        self.assertFalse(result.failures)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .plugins import ModelPluginBase
from anyblok.column import DateTime
from anyblok.mapper import ModelMapper
from datetime import datetime


class AutoUpdatePlugin(ModelPluginBase):

    def after_model_construction(self, base, namespace,
                                 transformation_properties):
        """Add the sqlalchemy event

        :param base: the Model class
        :param namespace: the namespace of the model
        :param transformation_properties: the properties of the model
        """
        namespaces = [namespace]
        namespaces.extend(list(base.__depends__))
        fields = []
        for ns in namespaces:
            for c in self.registry.get(ns).loaded_columns:
                f = self.registry.loaded_namespaces_first_step[namespace].get(c)
                if isinstance(f, DateTime) and f.auto_update:
                    fields.append(c)

        if fields:
            e = ModelMapper(namespace, 'after_update')

            def auto_update_listen(mapper, connection, target):
                now = datetime.now()
                for field in fields:
                    setattr(target, field, now)

            self.registry._sqlalchemy_known_events.append(
                (e, namespace, auto_update_listen))



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import DDLElement


class CreateView(DDLElement):
    def __init__(self, name, selectable):
        self.name = name
        self.selectable = selectable


class DropView(DDLElement):
    def __init__(self, name):
        self.name = name


@compiles(CreateView)
def compile_create_view(element, compiler, **kw):
    return "CREATE VIEW %s AS %s" % (
        element.name, compiler.sql_compiler.process(element.selectable))


@compiles(DropView)
def compile_drop_view(element, compiler, **kw):
    return "DROP VIEW IF EXISTS %s" % (element.name)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Denis VIVIÈS <dvivies@geoblink.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.bloks.anyblok_core.exceptions import ParameterException


@pytest.mark.usefixtures('rollback_registry')
class TestSystemParameter:

    def test_set(self, rollback_registry):
        registry = rollback_registry
        Parameter = registry.System.Parameter
        query = Parameter.query().filter(Parameter.key == 'test.parameter')
        if query.count():
            pytest.fail('key for test already existing')

        Parameter.set('test.parameter', True)
        assert query.count() == 1
        assert query.first().value == {'value': True}
        assert query.first().multi is False

    def test_set_with_multi(self, rollback_registry):
        registry = rollback_registry
        Parameter = registry.System.Parameter
        query = Parameter.query().filter(Parameter.key == 'test.parameter')
        if query.count():
            pytest.fail('key for test already existing')

        Parameter.set('test.parameter', {'test': True})
        assert query.count() == 1
        assert query.first().value == {'test': True}
        assert query.first().multi is True

    def test_get(self, rollback_registry):
        registry = rollback_registry
        Parameter = registry.System.Parameter
        Parameter.set('test.parameter', True)
        assert Parameter.get('test.parameter') is True

    def test_get_with_multi(self, rollback_registry):
        registry = rollback_registry
        Parameter = registry.System.Parameter
        Parameter.set('test.parameter', {'test': True})
        assert Parameter.get('test.parameter') == {'test': True}

    def test_unexisting_get(self, rollback_registry):
        registry = rollback_registry
        Parameter = registry.System.Parameter
        with pytest.raises(ParameterException):
            Parameter.get('test.parameter')

    def test_count(self, rollback_registry):
        registry = rollback_registry
        Parameter = registry.System.Parameter
        assert Parameter.is_exist('test.parameter') is False
        Parameter.set('test.parameter', True)
        assert Parameter.is_exist('test.parameter') is True

    def test_set_existing_key(self, rollback_registry):
        registry = rollback_registry
        Parameter = registry.System.Parameter
        query = Parameter.query().filter(Parameter.key == 'test.parameter')
        assert query.count() == 0
        Parameter.set('test.parameter', True)
        assert query.count() == 1
        assert Parameter.get('test.parameter') is True
        Parameter.set('test.parameter', False)
        assert query.count() == 1
        assert Parameter.get('test.parameter') is False



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.column import Integer
from anyblok.relationship import Many2Many, One2Many, Many2One
from .conftest import init_registry, reset_db


class TestInstrumentedListStdCase:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_blok):
        transaction = registry_blok.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_all_method_on_query_return_InstrumentedList(self, registry_blok):
        registry = registry_blok
        check = isinstance(registry.System.Blok.query().all(),
                           registry.InstrumentedList)
        assert check

    def test_emulates(self, registry_blok):
        registry = registry_blok
        assert not (hasattr(registry.InstrumentedList, '__emulates'))

    def test_empty_result_on_query_return_InstrumentedList(self, registry_blok):
        registry = registry_blok
        Blok = registry.System.Blok
        bloks = Blok.query().filter(Blok.name == 'Unexisting blok').all()
        check = isinstance(bloks, registry.InstrumentedList)
        assert check
        assert bloks.name == []


class TestInstrumentedList:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        reset_db()
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_M2M_with_InstrumentedList(self):

        def m2m_with_instrumentedlist():

            from anyblok import Declarations
            Model = Declarations.Model

            @Declarations.register(Model)
            class Test:
                id = Integer(primary_key=True)

            @Declarations.register(Model)
            class Test2:
                id = Integer(primary_key=True)
                tests = Many2Many(model=Model.Test, many2many="tests2")

        registry = self.init_registry(m2m_with_instrumentedlist)

        t = registry.Test.insert()
        t2 = registry.Test2.insert()
        t.tests2.append(t2)
        assert t2.tests == [t]
        check = isinstance(t.tests2, registry.InstrumentedList)
        assert check
        check = isinstance(t2.tests, registry.InstrumentedList)
        assert check

    def test_O2M_is_InstrumentedList(self):

        def o2m_with_instrumentedlist():

            from anyblok import Declarations
            Model = Declarations.Model

            @Declarations.register(Model)
            class Test2:
                id = Integer(primary_key=True)

            @Declarations.register(Model)
            class Test:
                id = Integer(primary_key=True)
                test2 = Integer(foreign_key=Model.Test2.use('id'))

            @Declarations.register(Model)  # noqa
            class Test2:
                tests = One2Many(model=Model.Test)

        registry = self.init_registry(o2m_with_instrumentedlist)

        t = registry.Test.insert()
        t2 = registry.Test2.insert()
        t2.tests.append(t)
        check = isinstance(t2.tests, registry.InstrumentedList)
        assert check

    def test_O2M_linked_is_InstrumentedList(self):

        def o2m_with_instrumentedlist():

            from anyblok import Declarations
            Model = Declarations.Model

            @Declarations.register(Model)
            class Test:
                id = Integer(primary_key=True)

            @Declarations.register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, one2many="tests2")

        registry = self.init_registry(o2m_with_instrumentedlist)

        t = registry.Test.insert()
        t2 = registry.Test2.insert()
        t.tests2.append(t2)
        check = isinstance(t.tests2, registry.InstrumentedList)
        assert check

    def test_call_column(self):

        def call_column():

            from anyblok import Declarations
            Model = Declarations.Model

            @Declarations.register(Model)
            class Test:
                id = Integer(primary_key=True)

        registry = self.init_registry(call_column)

        t = registry.Test.insert()
        assert registry.Test.query().all().id == [t.id]

    def test_call_method(self):

        def call_method():

            from anyblok import Declarations
            Model = Declarations.Model

            @Declarations.register(Model)
            class Test:
                id = Integer(primary_key=True)

                def foo(self):
                    return self.id

        registry = self.init_registry(call_method)

        t = registry.Test.insert()
        assert registry.Test.query().all().foo() == [t.id]

    def test_inherit(self):

        def inherit():

            from anyblok import Declarations
            Core = Declarations.Core

            @Declarations.register(Core)
            class InstrumentedList:

                def foo(self):
                    return True

        registry = self.init_registry(inherit)
        assert registry.System.Blok.query().all().foo()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.testing import sgdb_in
import pytest
from anyblok.config import Configuration
from anyblok.testing import tmp_configuration
from sqlalchemy import Integer as SA_Integer, String as SA_String
from sqlalchemy.exc import StatementError
from anyblok import Declarations
from anyblok.field import FieldException
from .conftest import init_registry, reset_db
from anyblok.column import (
    Column, Boolean, Json, String, BigInteger, Text, Selection, Date, DateTime,
    Time, Interval, Decimal, Float, LargeBinary, Integer, Sequence, Color,
    Password, UUID, URL, PhoneNumber, Email, Country)

try:
    import cryptography  # noqa
    has_cryptography = True
except Exception:
    has_cryptography = False

try:
    import passlib  # noqa
    has_passlib = True
except Exception:
    has_passlib = False

try:
    import colour  # noqa
    has_colour = True
except Exception:
    has_colour = False

try:
    import furl  # noqa
    has_furl = True
except Exception:
    has_furl = False


try:
    import phonenumbers  # noqa
    has_phonenumbers = True
except Exception:
    has_phonenumbers = False

try:
    import pycountry  # noqa
    has_pycountry = True
    from sqlalchemy_utils import PhoneNumber as PN
except Exception:
    has_pycountry = False


Model = Declarations.Model
register = Declarations.register


class OneColumn(Column):
    sqlalchemy_type = SA_Integer


class TestColumn:

    def test_forbid_instance(self):
        with pytest.raises(FieldException):
            Column()

    def test_without_label(self):
        column = OneColumn()
        column.get_sqlalchemy_mapping(None, None, 'a_column', None)
        assert column.label == 'A column'


def simple_column(ColumnType=None, **kwargs):

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        col = ColumnType(**kwargs)

        @classmethod
        def meth_secretkey(cls):
            return 'secretkey'


def column_with_foreign_key():

    @register(Model)
    class Test:

        name = String(primary_key=True)

    @register(Model)
    class Test2:

        id = Integer(primary_key=True)
        test = String(foreign_key=Model.Test.use('name'))


class TestColumns:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        reset_db()
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_column_with_type_in_kwargs(self):
        self.init_registry(
            simple_column, ColumnType=Integer, type_=Integer)

    def test_column_with_db_column_name_in_kwargs(self):
        registry = self.init_registry(simple_column, ColumnType=Integer,
                                      db_column_name='another_name')
        test = registry.Test.insert(col=1)
        assert test.col == 1
        res = registry.execute('select id from test where another_name=1')
        assert res.fetchone()[0] == test.id

    def test_column_with_foreign_key(self):
        registry = self.init_registry(column_with_foreign_key)
        registry.Test.insert(name='test')
        registry.Test2.insert(test='test')

    def test_integer(self):
        registry = self.init_registry(simple_column, ColumnType=Integer)
        test = registry.Test.insert(col=1)
        assert test.col == 1

    def test_integer_str_foreign_key(self):
        registry = self.init_registry(
            simple_column, ColumnType=Integer, foreign_key='Model.Test=>id')
        test = registry.Test.insert()
        test2 = registry.Test.insert(col=test.id)
        assert test2.col == test.id

    def test_big_integer(self):
        registry = self.init_registry(simple_column, ColumnType=BigInteger)
        test = registry.Test.insert(col=1)
        assert test.col == 1

    def test_Float(self):
        registry = self.init_registry(simple_column, ColumnType=Float)
        test = registry.Test.insert(col=1.0)
        assert test.col == 1.0

    def test_decimal(self):
        from decimal import Decimal as D

        registry = self.init_registry(simple_column, ColumnType=Decimal)
        test = registry.Test.insert(col=D('1.0'))
        assert test.col == D('1.0')

    def test_setter_decimal(self):
        from decimal import Decimal as D

        registry = self.init_registry(simple_column, ColumnType=Decimal)
        test = registry.Test.insert()
        test.col = '1.0'
        assert test.col == D('1.0')

    def test_boolean(self):
        registry = self.init_registry(simple_column, ColumnType=Boolean)
        test = registry.Test.insert(col=True)
        assert test.col

    def test_boolean_with_default(self):
        registry = self.init_registry(simple_column, ColumnType=Boolean,
                                      default=False)
        test = registry.Test.insert()
        assert test.col is False

    def test_string(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert(col='col')
        assert test.col == 'col'

    def test_string_with_False(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert(col=False)
        assert test.col is False
        self.registry.flush()
        self.registry.expire(test, ['col'])
        assert test.col == ''

    def test_string_set_False(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert()
        test.col = False
        assert test.col is False
        self.registry.flush()
        self.registry.expire(test, ['col'])
        assert test.col == ''

    def test_string_query_False(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert()
        self.registry.Test.query().filter_by(id=test.id).update({'col': False})
        assert test.col is False
        self.registry.expire(test, ['col'])
        assert test.col == ''

    @pytest.mark.skipif(not has_cryptography,
                        reason="cryptography is not installed")
    def test_string_with_encrypt_key(self):
        registry = self.init_registry(simple_column, ColumnType=String,
                                      encrypt_key='secretkey')
        test = registry.Test.insert(col='col')
        registry.session.commit()
        assert test.col == 'col'
        res = registry.execute('select col from test where id = %s' % test.id)
        res = res.fetchall()[0][0]
        assert res != 'col'

    @pytest.mark.skipif(not has_cryptography,
                        reason="cryptography is not installed")
    def test_string_with_encrypt_key_defined_by_a_method(self):
        registry = self.init_registry(simple_column, ColumnType=String,
                                      encrypt_key='meth_secretkey')
        test = registry.Test.insert(col='col')
        registry.session.commit()
        assert test.col == 'col'
        res = registry.execute('select col from test where id = %s' % test.id)
        res = res.fetchall()[0][0]
        assert res != 'col'

    @pytest.mark.skipif(not has_cryptography,
                        reason="cryptography is not installed")
    def test_string_with_encrypt_key_defined_by_configuration(self):
        Configuration.set('default_encrypt_key', 'secretkey')
        registry = self.init_registry(simple_column, ColumnType=String,
                                      encrypt_key=True)
        test = registry.Test.insert(col='col')
        registry.session.commit()
        assert test.col == 'col'
        res = registry.execute('select col from test where id = %s' % test.id)
        res = res.fetchall()[0][0]
        assert res != 'col'
        del Configuration.configuration['default_encrypt_key']

    @pytest.mark.skipif(not has_cryptography,
                        reason="cryptography is not installed")
    def test_datetime_with_encrypt_key(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime,
                                      encrypt_key='secretkey')
        test = registry.Test.insert(col=now)
        registry.session.commit()
        assert test.col == now

    def test_string_with_size(self):
        registry = self.init_registry(
            simple_column, ColumnType=String, size=100)
        test = registry.Test.insert(col='col')
        assert test.col == 'col'

    @pytest.mark.skipif(not has_passlib,
                        reason="passlib is not installed")
    def test_password(self):
        registry = self.init_registry(simple_column, ColumnType=Password,
                                      crypt_context={'schemes': ['md5_crypt']})
        test = registry.Test.insert(col='col')
        assert test.col == 'col'
        assert registry.execute('Select col from test').fetchone()[0] != 'col'

    def test_text(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert(col='col')
        assert test.col == 'col'

    def test_text_with_False(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert(col=False)
        assert test.col is False
        self.registry.flush()
        self.registry.expire(test, ['col'])
        assert test.col == ''

    def test_text_set_False(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert()
        test.col = False
        assert test.col is False
        self.registry.flush()
        self.registry.expire(test, ['col'])
        assert test.col == ''

    def test_text_query_False(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert()
        self.registry.Test.query().filter_by(id=test.id).update({'col': False})
        assert test.col is False
        self.registry.expire(test, ['col'])
        assert test.col == ''

    def test_date(self):
        from datetime import date

        now = date.today()
        registry = self.init_registry(simple_column, ColumnType=Date)
        test = registry.Test.insert(col=now)
        assert test.col == now

    def test_datetime(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        assert test.col == now

    def test_datetime_none_value(self):
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=None)
        assert test.col is None

    def test_datetime_str_conversion_1(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%z'))
        assert test.col == now

    def test_datetime_str_conversion_2(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%Z'))
        assert test.col == now

    def test_datetime_str_conversion_3(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S.%f'))
        assert test.col == now

    def test_datetime_str_conversion_4(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S'))
        assert test.col == now.replace(microsecond=0)

    def test_datetime_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now
        assert test.col == now

    def test_datetime_by_property_none_value(self):
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = None
        assert test.col is None

    def test_datetime_str_conversion_1_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S.%f%z')
        assert test.col == now

    def test_datetime_str_conversion_2_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S.%f%Z')
        assert test.col == now

    def test_datetime_str_conversion_3_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S.%f')
        assert test.col == now

    def test_datetime_str_conversion_4_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S')
        assert test.col == now.replace(microsecond=0)

    def test_datetime_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(dict(col=now))
        assert test.col == now

    def test_datetime_by_query_none_value(self):
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(dict(col=None))
        assert test.col is None

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #87')
    def test_datetime_str_conversion_1_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%z')))
        registry.expire(test, ['col'])
        assert test.col == now

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #87')
    def test_datetime_str_conversion_2_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%Z')))
        registry.expire(test, ['col'])
        assert test.col == now

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #87')
    def test_datetime_str_conversion_3_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S.%f')))
        registry.expire(test, ['col'])
        assert test.col == now

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #87')
    def test_datetime_str_conversion_4_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S')))
        registry.expire(test, ['col'])
        assert test.col == now.replace(microsecond=0)

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #87')
    def test_datetime_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        assert Test.query().filter(Test.col == now).one() is test

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #87')
    def test_datetime_str_conversion_1_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        assert Test.query().filter(
            Test.col == now.strftime('%Y-%m-%d %H:%M:%S.%f%z')).one() is test

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #87')
    def test_datetime_str_conversion_2_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        assert Test.query().filter(
            Test.col == now.strftime('%Y-%m-%d %H:%M:%S.%f%Z')).one() is test

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #87')
    def test_datetime_str_conversion_3_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        assert Test.query().filter(
            Test.col == now.strftime('%Y-%m-%d %H:%M:%S.%f')).one() is test

    def test_datetime_without_auto_update_1(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:
                id = Integer(primary_key=True)
                update_at = DateTime()
                val = String()

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(val='first add')
        assert test.update_at is None
        test.val = 'other'
        registry.flush()
        assert test.update_at is None

    def test_datetime_without_auto_update_2(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:
                id = Integer(primary_key=True)
                update_at = DateTime(auto_update=False)
                val = String()

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(val='first add')
        assert test.update_at is None
        test.val = 'other'
        registry.flush()
        assert test.update_at is None

    def test_datetime_with_auto_update(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:
                id = Integer(primary_key=True)
                update_at = DateTime(auto_update=True)
                val = String()

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(val='first add')
        assert test.update_at is None
        test.val = 'other'
        registry.flush()
        assert test.update_at is not None

    def test_datetime_with_default_timezone_tz(self):
        import datetime
        import pytz

        timezone = pytz.timezone('Asia/Tokyo')
        now = datetime.datetime.now()
        registry = self.init_registry(
            simple_column, ColumnType=DateTime,
            default_timezone=timezone)
        field = registry.loaded_namespaces_first_step['Model.Test']['col']
        assert field.default_timezone is timezone

        test = registry.Test.insert(col=now)
        assert test.col.tzinfo.zone is timezone.zone

    def test_datetime_with_default_timezone_str(self):
        import datetime
        import pytz

        timezone = pytz.timezone('Asia/Tokyo')
        now = datetime.datetime.now()
        registry = self.init_registry(
            simple_column, ColumnType=DateTime,
            default_timezone='Asia/Tokyo')
        field = registry.loaded_namespaces_first_step['Model.Test']['col']
        assert field.default_timezone == timezone

        test = registry.Test.insert(col=now)
        assert test.col.tzinfo.zone == timezone.zone

    def test_datetime_with_default_global_timezone_str(self):
        import datetime
        import pytz

        timezone = pytz.timezone('Asia/Tokyo')
        now = datetime.datetime.now()
        with tmp_configuration(default_timezone='Asia/Tokyo'):
            registry = self.init_registry(simple_column, ColumnType=DateTime)

        field = registry.loaded_namespaces_first_step['Model.Test']['col']
        assert field.default_timezone is timezone

        test = registry.Test.insert(col=now)
        assert test.col.tzinfo.zone is timezone.zone

    def test_interval(self):
        from datetime import timedelta

        dt = timedelta(days=5)
        registry = self.init_registry(simple_column, ColumnType=Interval)
        test = registry.Test.insert(col=dt)
        assert test.col == dt

    def test_time(self):
        from datetime import time

        now = time()
        registry = self.init_registry(simple_column, ColumnType=Time)
        test = registry.Test.insert(col=now)
        assert test.col == now

    def test_large_binary(self):
        from os import urandom

        blob = urandom(10000)
        registry = self.init_registry(simple_column, ColumnType=LargeBinary)

        test = registry.Test.insert(col=blob)
        assert test.col == blob

    def test_selection(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        test = registry.Test.insert(col=SELECTIONS[0][0])
        assert test.col == SELECTIONS[0][0]
        assert test.col.label == SELECTIONS[0][1]
        test = registry.Test.query().first()
        assert test.col == SELECTIONS[0][0]
        assert test.col.label == SELECTIONS[0][1]
        with pytest.raises(FieldException):
            test.col = 'bad value'

    def test_selection_with_only_one_value(self):
        SELECTIONS = [
            ('admin', 'Admin'),
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        test = registry.Test.insert(col=SELECTIONS[0][0])
        assert test.col == SELECTIONS[0][0]
        assert test.col.label == SELECTIONS[0][1]
        test = registry.Test.query().first()
        assert test.col == SELECTIONS[0][0]
        assert test.col.label == SELECTIONS[0][1]
        with pytest.raises(FieldException):
            test.col = 'bad value'

    def test_selection_without_value(self):
        self.init_registry(simple_column, ColumnType=Selection, selections=[])

    def test_selection_autodoc(self):
        SELECTIONS = [
            ('admin', 'Admin'),
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        column = registry.System.Column.query().filter_by(
            model='Model.Test', name='col').one()
        assert column._description() == {
                'id': 'col',
                'label': 'Col',
                'model': None,
                'nullable': True,
                'primary_key': False,
                'selections': [
                    ('admin', 'Admin'),
                ],
                'type': 'Selection',
            }

    def test_selection_with_none_value(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        t = registry.Test.insert()
        assert t.col is None

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #90')
    def test_selection_change_by_query(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        registry.Test.insert(col=SELECTIONS[0][0])
        with pytest.raises(StatementError):
            registry.Test.query().update({'col': 'bad value'})

    def test_selection_like_comparator(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        Test = registry.Test
        t = Test.insert(col=SELECTIONS[0][0])
        t1 = Test.query().filter(Test.col.like('%admin%')).one()
        assert t is t1

    def test_selection_key_other_than_str(self):
        SELECTIONS = [
            (0, 'Admin'),
            (1, 'Regular user')
        ]

        with pytest.raises(FieldException):
            self.init_registry(
                simple_column, ColumnType=Selection, selections=SELECTIONS)

    def test_selection_comparator(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        registry.Test.insert(col=SELECTIONS[0][0])
        registry.Test.query().filter(
            registry.Test.col.in_(['admin', 'regular-user'])).first()

    def test_selection_use_method(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        def add_selection():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                col = Selection(selections='get_selection')

                @classmethod
                def get_selection(cls):
                    return SELECTIONS

        registry = self.init_registry(add_selection)
        registry.Test.insert(col=SELECTIONS[0][0])
        registry.Test.query().filter(
            registry.Test.col.in_(['admin', 'regular-user'])).first()

    @pytest.mark.skipif(sgdb_in(['MariaDB']),
                        reason='JSON is not existing in this SGDB')
    def test_json(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        val = {'a': 'Test'}
        test = registry.Test.insert(col=val)
        assert test.col == val

    @pytest.mark.skipif(sgdb_in(['MariaDB']),
                        reason='JSON is not existing in this SGDB')
    def test_json_update(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        test = registry.Test.insert(col={'a': 'test'})
        test.col['b'] = 'test'
        assert test.col == {'a': 'test', 'b': 'test'}

    @pytest.mark.skipif(sgdb_in(['MariaDB']),
                        reason='JSON is not existing in this SGDB')
    def test_json_simple_filter(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        Test = registry.Test
        Test.insert(col={'a': 'test'})
        Test.insert(col={'a': 'test'})
        Test.insert(col={'b': 'test'})
        assert Test.query().filter(
            Test.col['a'].cast(SA_String) == '"test"').count() == 2

    @pytest.mark.skipif(sgdb_in(['MariaDB']),
                        reason='JSON is not existing in this SGDB')
    def test_json_null(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        Test = registry.Test
        Test.insert(col=None)
        Test.insert(col=None)
        Test.insert(col={'a': 'test'})
        assert Test.query().filter(Test.col.is_(None)).count() == 2
        assert Test.query().filter(Test.col.isnot(None)).count() == 1

    def test_add_default_classmethod(self):
        val = 'test'

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default="get_val")

                @classmethod
                def get_val(cls):
                    return val

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        assert t.val == val

    def test_add_default_without_classmethod(self):
        value = 'test'

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default="get_val")

                def get_val(cls):
                    return value

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        assert t.val == "get_val"

    def test_add_default_by_var(self):
        value = 'test'

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default=value)

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        assert t.val == value

    def test_add_default(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default='value')

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        assert t.val == 'value'

    def test_add_field_as_default(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default='val')

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        assert t.val == 'val'

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #89')
    def test_sequence(self):
        registry = self.init_registry(simple_column, ColumnType=Sequence)
        assert registry.Test.insert().col == "1"
        assert registry.Test.insert().col == "2"
        assert registry.Test.insert().col == "3"
        assert registry.Test.insert().col == "4"
        Seq = registry.System.Sequence
        assert Seq.query().filter(Seq.code == 'Model.Test=>col').count() == 1

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #89')
    def test_sequence_with_primary_key(self):
        registry = self.init_registry(simple_column, ColumnType=Sequence,
                                      primary_key=True)
        assert registry.Test.insert().col == "1"
        assert registry.Test.insert().col == "2"

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #89')
    def test_sequence_with_code_and_formater(self):
        registry = self.init_registry(simple_column, ColumnType=Sequence,
                                      code="SO", formater="{code}-{seq:06d}")
        assert registry.Test.insert().col == "SO-000001"
        assert registry.Test.insert().col == "SO-000002"
        assert registry.Test.insert().col == "SO-000003"
        assert registry.Test.insert().col == "SO-000004"
        Seq = registry.System.Sequence
        assert Seq.query().filter(Seq.code == 'SO').count() == 1

    def test_sequence_with_foreign_key(self):
        with pytest.raises(FieldException):
            self.init_registry(simple_column, ColumnType=Sequence,
                               foreign_key=Model.System.Model.use('name'))

    def test_sequence_with_default(self):
        with pytest.raises(FieldException):
            self.init_registry(simple_column, ColumnType=Sequence,
                               default='default value')

    @pytest.mark.skipif(not has_colour,
                        reason="colour is not installed")
    def test_color(self):
        color = '#F5F5F5'
        registry = self.init_registry(simple_column, ColumnType=Color)
        test = registry.Test.insert(col=color)
        assert test.col.hex == colour.Color(color).hex

    @pytest.mark.skipif(not has_colour,
                        reason="colour is not installed")
    def test_setter_color(self):
        color = '#F5F5F5'
        registry = self.init_registry(simple_column, ColumnType=Color)
        test = registry.Test.insert()
        test.col = color
        assert test.col.hex == colour.Color(color).hex

    def test_uuid_binary_1(self):
        from uuid import uuid1
        uuid = uuid1()
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        assert test.col is uuid

    def test_uuid_binary_3(self):
        from uuid import uuid3, NAMESPACE_DNS
        uuid = uuid3(NAMESPACE_DNS, 'python.org')
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        assert test.col is uuid

    def test_uuid_binary_4(self):
        from uuid import uuid4
        uuid = uuid4()
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        assert test.col is uuid

    def test_uuid_binary_5(self):
        from uuid import uuid5, NAMESPACE_DNS
        uuid = uuid5(NAMESPACE_DNS, 'python.org')
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        assert test.col is uuid

    def test_uuid_char32(self):
        from uuid import uuid1
        uuid = uuid1()
        registry = self.init_registry(simple_column, ColumnType=UUID,
                                      binary=False)
        test = registry.Test.insert(col=uuid)
        assert test.col is uuid

    @pytest.mark.skipif(not has_furl, reason="furl is not installed")
    def test_URL(self):
        f = furl.furl('http://doc.anyblok.org')
        registry = self.init_registry(simple_column, ColumnType=URL)
        test = registry.Test.insert(col=f)
        assert test.col == f

    @pytest.mark.skipif(not has_furl, reason="furl is not installed")
    def test_setter_URL(self):
        f = 'http://doc.anyblok.org'
        registry = self.init_registry(simple_column, ColumnType=URL)
        test = registry.Test.insert()
        test.col = f
        assert test.col.url == f

    @pytest.mark.skipif(not has_furl, reason="furl is not installed")
    def test_URL_2(self):
        f = 'http://doc.anyblok.org'
        registry = self.init_registry(simple_column, ColumnType=URL)
        test = registry.Test.insert(col=f)
        assert test.col.url == f

    @pytest.mark.skipif(not has_furl, reason="furl is not installed")
    def test_URL_3(self):
        f = 'http://doc.anyblok.org'
        registry = self.init_registry(simple_column, ColumnType=URL)
        registry.Test.insert(col=f)
        Test = registry.Test
        test = Test.query().filter(Test.col == f).one()
        assert test.col.url == f

    @pytest.mark.skipif(not has_furl, reason="furl is not installed")
    def test_URL_4(self):
        f = furl.furl('http://doc.anyblok.org')
        registry = self.init_registry(simple_column, ColumnType=URL)
        registry.Test.insert(col=f)
        Test = registry.Test
        test = Test.query().filter(Test.col == f).one()
        assert test.col.url == f.url

    @pytest.mark.skipif(not has_phonenumbers,
                        reason="phonenumbers is not installed")
    def test_phonenumbers_at_insert(self):
        registry = self.init_registry(simple_column, ColumnType=PhoneNumber)
        test = registry.Test.insert(col='+120012301')
        assert test.col.national == '20012301'
        getted = registry.execute('Select col from test').fetchone()[0]
        assert getted == '+120012301'

    @pytest.mark.skipif(not has_phonenumbers,
                        reason="phonenumbers is not installed")
    def test_phonenumbers_at_setter(self):
        registry = self.init_registry(simple_column, ColumnType=PhoneNumber)
        test = registry.Test.insert()
        test.col = '+120012301'
        assert test.col.national == '20012301'
        registry.flush()
        getted = registry.execute('Select col from test').fetchone()[0]
        assert getted == '+120012301'

    @pytest.mark.skipif(not has_phonenumbers,
                        reason="phonenumbers is not installed")
    def test_phonenumbers_obj_at_insert(self):
        registry = self.init_registry(simple_column, ColumnType=PhoneNumber)
        col = PN("+120012301", None)
        test = registry.Test.insert(col=col)
        assert test.col == col
        getted = registry.execute('Select col from test').fetchone()[0]
        assert getted == '+120012301'

    @pytest.mark.skipif(not has_phonenumbers,
                        reason="phonenumbers is not installed")
    def test_phonenumbers_obj_at_setter(self):
        registry = self.init_registry(simple_column, ColumnType=PhoneNumber)
        test = registry.Test.insert()
        test.col = PN("+120012301", None)
        assert test.col.national == '20012301'
        registry.flush()
        getted = registry.execute('Select col from test').fetchone()[0]
        assert getted == '+120012301'

    @pytest.mark.skipif(not has_phonenumbers,
                        reason="phonenumbers is not installed")
    def test_phonenumbers_obj_at_setter_with_empty_value(self):
        registry = self.init_registry(simple_column, ColumnType=PhoneNumber)
        test = registry.Test.insert()
        test.col = ""
        assert test.col == ''
        registry.flush()
        assert registry.execute('Select col from test').fetchone()[0] == ''

    def test_email_at_insert(self):
        registry = self.init_registry(simple_column, ColumnType=Email)
        test = registry.Test.insert(col='John.Smith@foo.com')
        assert test.col == 'john.smith@foo.com'
        getted = registry.Test.query().filter_by(
            col='John.Smith@foo.com').count()
        assert getted == 1

    def test_email_at_setter(self):
        registry = self.init_registry(simple_column, ColumnType=Email)
        test = registry.Test.insert()
        test.col = 'John.Smith@foo.com'
        assert test.col == 'john.smith@foo.com'
        assert registry.Test.query().filter_by(
            col='John.Smith@foo.com').count() == 1

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_at_insert(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        test = registry.Test.insert(col='FR')
        assert test.col is pycountry.countries.get(alpha_2='FR')
        assert test.col.name == 'France'
        assert registry.execute('Select col from test').fetchone()[0] == 'FRA'

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_at_insert_with_alpha_3(self):
        registry = self.init_registry(
            simple_column, ColumnType=Country, mode='alpha_3')
        test = registry.Test.insert(col='FRA')
        assert test.col is pycountry.countries.get(alpha_2='FR')
        assert test.col.name == 'France'
        assert registry.execute('Select col from test').fetchone()[0] == 'FRA'

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_at_insert_with_object(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        fr = pycountry.countries.get(alpha_2='FR')
        test = registry.Test.insert(col=fr)
        assert test.col is fr
        assert test.col.name == 'France'
        assert registry.execute('Select col from test').fetchone()[0] == 'FRA'

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_at_update(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        test = registry.Test.insert()
        test.col = 'FR'
        registry.flush()
        assert test.col is pycountry.countries.get(alpha_2='FR')
        assert test.col.name == 'France'
        assert registry.execute('Select col from test').fetchone()[0] == 'FRA'

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_at_update_with_alpha_3(self):
        registry = self.init_registry(
            simple_column, ColumnType=Country, mode='alpha_3')
        test = registry.Test.insert()
        test.col = 'FRA'
        registry.flush()
        assert test.col is pycountry.countries.get(alpha_2='FR')
        assert test.col.name == 'France'
        assert registry.execute('Select col from test').fetchone()[0] == 'FRA'

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_at_update_with_object(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        fr = pycountry.countries.get(alpha_2='FR')
        test = registry.Test.insert()
        test.col = fr
        registry.flush()
        assert test.col is fr
        assert test.col.name == 'France'
        assert registry.execute('Select col from test').fetchone()[0] == 'FRA'

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_query_is_with_object(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        Test = registry.Test
        fr = pycountry.countries.get(alpha_2='FR')
        Test.insert(col='FR')
        assert Test.query().filter(Test.col == fr).count() == 1

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_query_is_with_alpha_3(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        Test = registry.Test
        Test.insert(col='FR')
        assert Test.query().filter(Test.col == 'FRA').count() == 1

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_query_insert_wrong(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        with pytest.raises(LookupError):
            registry.Test.insert(col='WG')

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #90')
    def test_pycoundtry_query_insert_by_wrong_query(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        with pytest.raises(Exception):
            registry.execute("insert into test (col) values ('WG2')")


class TestColumnsAutoDoc:

    def call_autodoc(self, column, **kwargs):
        col = column(**kwargs)
        col.autodoc()

    def test_integer(self):
        self.call_autodoc(Integer)

    def test_big_integer(self):
        self.call_autodoc(BigInteger)

    def test_Float(self):
        self.call_autodoc(Float)

    def test_decimal(self):
        self.call_autodoc(Decimal)

    def test_boolean(self):
        self.call_autodoc(Boolean)

    def test_string(self):
        self.call_autodoc(String)

    @pytest.mark.skipif(not has_cryptography,
                        reason="cryptography is not installed")
    def test_string_with_encrypt_key(self):
        self.call_autodoc(String, encrypt_key='secretkey')

    @pytest.mark.skipif(not has_cryptography,
                        reason="cryptography is not installed")
    def test_datetime_with_encrypt_key(self):
        self.call_autodoc(DateTime, encrypt_key='secretkey')

    def test_string_with_size(self):
        self.call_autodoc(String, size=100)

    @pytest.mark.skipif(not has_passlib, reason="passlib is not installed")
    def test_password(self):
        self.call_autodoc(Password, crypt_context={'schemes': ['md5_crypt']})

    def test_text(self):
        self.call_autodoc(Text)

    def test_date(self):
        self.call_autodoc(Date)

    def test_datetime(self):
        self.call_autodoc(DateTime)

    def test_datetime_with_default_timezone_tz(self):
        import pytz
        timezone = pytz.timezone('Asia/Tokyo')
        self.call_autodoc(DateTime, default_timezone=timezone)

    def test_datetime_with_default_timezone_str(self):
        self.call_autodoc(DateTime, default_timezone='Asia/Tokyo')

    def test_interval(self):
        self.call_autodoc(Interval)

    def test_time(self):
        self.call_autodoc(Time)

    def test_large_binary(self):
        self.call_autodoc(LargeBinary)

    def test_selection(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]
        self.call_autodoc(Selection, selections=SELECTIONS)

    def test_json(self):
        self.call_autodoc(Json)

    def test_add_default(self):
        self.call_autodoc(String, default='get_val')

    def test_add_default_by_var(self):
        value = 'test'
        self.call_autodoc(String, default=value)

    def test_sequence(self):
        self.call_autodoc(Sequence)

    def test_sequence_with_primary_key(self):
        self.call_autodoc(Sequence, primary_key=True)

    @pytest.mark.skipif(not has_colour, reason="colour is not installed")
    def test_color(self):
        self.call_autodoc(Color)

    def test_uuid_binary_1(self):
        self.call_autodoc(UUID)

    def test_uuid_char32(self):
        self.call_autodoc(UUID, binary=False)

    @pytest.mark.skipif(not has_furl, reason="furl is not installed")
    def test_URL(self):
        self.call_autodoc(URL)

    @pytest.mark.skipif(not has_phonenumbers,
                        reason="phonenumbers is not installed")
    def test_phonenumbers_at_insert(self):
        self.call_autodoc(PhoneNumber)

    def test_email_at_insert(self):
        self.call_autodoc(Email)

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_at_insert(self):
        self.call_autodoc(Country)

    @pytest.mark.skipif(not has_pycountry, reason="pycountry is not installed")
    def test_pycoundtry_at_insert_with_alpha_3(self):
        self.call_autodoc(Country, mode='alpha_3')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .factory import has_sql_fields  # noqa
from .factory import ModelFactory, ViewFactory, ModelFactoryException
import warnings

MODEL = 'MODEL'
VIEW = 'VIEW'


def get_factory(type_):
    warnings.warn(
        "type attribute is deprecated to define a type"
        " of model, you have to use the factory",
        DeprecationWarning, stacklevel=2)
    if type_ == MODEL:
        return ModelFactory
    elif type_ == VIEW:
        return ViewFactory
    else:
        raise ModelFactoryException("The type %r is unknown" % type_)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pkg_resources import iter_entry_points
from anyblok.imp import ImportManager
from .logging import log
from anyblok.environment import EnvironmentManager
from time import sleep
from sys import modules
from os.path import dirname
from logging import getLogger

logger = getLogger(__name__)


class BlokManagerException(LookupError):
    """ Simple exception to BlokManager """

    def __init__(self, *args, **kwargs):
        EnvironmentManager.set('current_blok', None)
        super(BlokManagerException, self).__init__(*args, **kwargs)


class BlokManager:
    """ Manage the bloks for one process

    A blok has a `setuptools` entrypoint, this entry point is defined
    by the ``entry_points`` attribute in the first load

    The ``bloks`` attribute is a dict with all the loaded entry points

    Use this class to import all the bloks in the entrypoint::

        BlokManager.load()

    """

    bloks = {}
    entry_points = None
    ordered_bloks = []
    auto_install = []

    @classmethod
    def list(cls):
        """ Return the ordered bloks

        :rtype: list of blok name ordered by loading
        """
        return cls.ordered_bloks

    @classmethod
    def has(cls, blok):
        """ Return True if the blok is loaded

        :param blok: blok name
        :rtype: bool
        """
        return blok and blok in cls.ordered_bloks or False

    @classmethod
    def get(cls, blok):
        """ Return the loaded blok

        :param blok: blok name
        :rtype: blok instance
        :exception: BlokManagerException
        """
        if not cls.has(blok):
            raise BlokManagerException('%r not found' % blok)

        return cls.bloks[blok]

    @classmethod
    def set(cls, blokname, blok):
        """ Add a new blok

        :param blokname: blok name
        :param blok: blok instance
        :exception: BlokManagerException
        """
        if cls.has(blokname):
            raise BlokManagerException('%r already present' % blokname)

        cls.bloks[blokname] = blok
        cls.ordered_bloks.append(blokname)

    @classmethod
    @log(logger, level='debug')
    def reload(cls):
        """ Reload the entry points

        Empty the ``bloks`` dict and use the ``entry_points`` attribute to
        load bloks
        :exception: BlokManagerException
        """
        if cls.entry_points is None:
            raise BlokManagerException(
                """You must use the ``load`` classmethod before using """
                """``reload``""")

        entry_points = []
        entry_points += cls.entry_points
        cls.unload()
        cls.load(entry_points=entry_points)

    @classmethod
    @log(logger, level='debug')
    def unload(cls):
        """ Unload all the bloks but not the registry """
        cls.bloks = {}
        cls.ordered_bloks = []
        cls.entry_points = None
        cls.auto_install = []
        from .registry import RegistryManager
        RegistryManager.unload()

    @classmethod
    def get_need_blok_linked_bloks(cls, blok):
        for required in cls.bloks[blok].required:
            if not cls.get_need_blok(required):
                raise BlokManagerException(
                    "Not %s required bloks found" % required)

            cls.bloks[required].required_by.append(blok)

        for optional in cls.bloks[blok].optional:
            if cls.get_need_blok(optional):
                cls.bloks[optional].optional_by.append(blok)

        for conditional in cls.bloks[blok].conditional:
            cls.bloks[conditional].conditional_by.append(blok)

        for conflicting in cls.bloks[blok].conflicting:
            cls.bloks[conflicting].conflicting_by.append(blok)

    @classmethod
    def get_need_blok(cls, blok):
        if cls.has(blok):
            return True

        if blok not in cls.bloks:
            return False

        cls.get_need_blok_linked_bloks(blok)
        cls.ordered_bloks.append(blok)
        EnvironmentManager.set('current_blok', blok)

        if not ImportManager.has(blok):
            # Import only if not exist don't reload here
            mod = ImportManager.add(blok)
            mod.imports()
        else:
            mod = ImportManager.get(blok)
            mod.reload()

        if cls.bloks[blok].autoinstall:
            cls.auto_install.append(blok)

        return True

    @classmethod
    @log(logger, level='debug')
    def load(cls, entry_points=('bloks',)):
        """ Load all the bloks and import them

        :param entry_points: Use by ``iter_entry_points`` to get the blok
        :exception: BlokManagerException
        """
        if not entry_points:
            raise BlokManagerException("The entry_points mustn't be empty")

        cls.entry_points = entry_points

        if EnvironmentManager.get('current_blok'):
            while EnvironmentManager.get('current_blok'):
                sleep(0.1)

        EnvironmentManager.set('current_blok', 'start')

        bloks = []
        for entry_point in entry_points:
            count = 0
            for i in iter_entry_points(entry_point):
                count += 1
                blok = i.load()
                blok.required_by = []
                blok.optional_by = []
                blok.conditional_by = []
                blok.conflicting_by = []
                cls.set(i.name, blok)
                blok.name = i.name
                bloks.append((blok.priority, i.name))

            if not count:
                raise BlokManagerException(
                    "Invalid bloks group %r" % entry_point)

        # Empty the ordered blok to reload it depending on the priority
        cls.ordered_bloks = []
        bloks.sort()

        try:
            while bloks:
                blok = bloks.pop(0)[1]
                cls.get_need_blok(blok)

        finally:
            EnvironmentManager.set('current_blok', None)

    @classmethod
    def getPath(cls, blok):
        """ Return the path of the blok

        :param blok: blok name in ``ordered_bloks``
        :rtype: absolute path
        """
        blok = cls.get(blok)
        return dirname(modules[blok.__module__].__file__)


class Blok:
    """ Super class for all the bloks

    define the default value for:

    * priority: order to load the blok
    * required: list of the bloks needed to install this blok
    * optional: list of the bloks to be installed if present in the blok list
    * conditional: if all the bloks of this list are installed then install
      this blok
    """

    autoinstall = False
    priority = 100
    required = []
    optional = []
    conditional = []
    conflicting = []
    name = None  # filled by the BlokManager
    author = ''
    logo = ''

    def __init__(self, registry):
        self.registry = registry

    @classmethod
    def import_declaration_module(cls):
        """ Do the python import for the Declaration of the model or other
        """

    def update(self, latest_version):
        """ Call at the installation or update

        :param latest_version: latest version installed, if the blok have not
                               been installing the latest_version will be None
        """

    def pre_migration(self, latest_version):
        """Call at update, before the automigration

        .. warning::

            You can not use the ORM

        :param latest_version: latest version installed, if the blok have not
                               been installing the latest_version will be None
        """

    def post_migration(self, latest_version):
        """Call at update, after the automigration

        :param latest_version: latest version installed, if the blok have not
                               been installing the latest_version will be None
        """

    def uninstall(self):
        """ Call at the uninstallation
        """

    def load(self):
        """ Call at the launch of the application
        """



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.blok import BlokManager, Blok, BlokManagerException


class TestBlokManager:

    @pytest.fixture(autouse=True)
    def bloks_must_be_unloaded(self, request):
        request.addfinalizer(BlokManager.unload)
        BlokManager.unload()

    def test_load_anyblok(self):
        BlokManager.load()
        if not BlokManager.list():
            pytest.fail('No blok load')
        if not BlokManager.has('anyblok-core'):
            pytest.fail("The blok 'anyblok-core' is missing")

        BlokManager.get('anyblok-core')

    def test_load_with_invalid_blok_group(self):
        with pytest.raises(BlokManagerException):
            BlokManager.load(entry_points=('Invalid blok group',))

    def test_load_without_blok_group(self):
        with pytest.raises(BlokManagerException):
            BlokManager.load(entry_points=())

    def test_reload(self):
        BlokManager.load()
        BlokManager.set('invalid blok', None)
        BlokManager.get('invalid blok')
        BlokManager.reload()
        with pytest.raises(BlokManagerException):
            BlokManager.get('invalid blok')

    def test_reload_without_load(self):
        with pytest.raises(BlokManagerException):
            BlokManager.reload()

    def test_get_invalid_blok(self):
        BlokManager.load()
        with pytest.raises(BlokManagerException):
            BlokManager.get('invalid blok')

    def test_set(self):
        blok_name = 'ABlok'
        BlokManager.set(blok_name, Blok)
        assert BlokManager.has(blok_name)

    def test_set_two_time(self):
        blok_name = 'ABlok'
        BlokManager.set(blok_name, Blok)
        with pytest.raises(BlokManagerException):
            BlokManager.set(blok_name, Blok)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Simon ANDRÉ <sandre@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok import Declarations
from anyblok.column import Integer, String, Date, DateTime
from anyblok.relationship import Many2One
from anyblok.field import Function
from datetime import date
from .conftest import init_registry


register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin
Core = Declarations.Core


def check_registry(Model, **kwargs):
    t = Model.insert(name="test", **kwargs)
    t2 = Model.query().filter(
        Model.type_entity == Model.__mapper__.polymorphic_identity).first()
    assert t2 is t
    return t


def multi_table_poly():
    @register(Model)
    class Employee:
        id = Integer(primary_key=True)
        name = String()
        type_entity = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Employee, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'employee',
                'polymorphic_on': cls.type_entity,
            })
            return mapper_args

    @register(Model)
    class Engineer(Model.Employee):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        engineer_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Engineer, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'engineer',
            })
            return mapper_args

    @register(Model)
    class Manager(Model.Employee):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        manager_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Manager, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'manager',
            })
            return mapper_args


@pytest.fixture(scope="class")
def registry_multi_table_poly(request, bloks_loaded):
    registry = init_registry(multi_table_poly)
    request.addfinalizer(registry.close)
    return registry


class TestMultiTablePoly:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_multi_table_poly):
        transaction = registry_multi_table_poly.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_field_description(self, registry_multi_table_poly):
        registry = registry_multi_table_poly
        fd_employee = list(registry.Employee.fields_description().keys())
        fd_employee.sort()
        fd_engineer = list(registry.Engineer.fields_description().keys())
        fd_engineer.sort()
        fd_manager = list(registry.Manager.fields_description().keys())
        fd_manager.sort()
        assert fd_employee == ['id', 'name', 'type_entity']
        assert fd_engineer == ['engineer_name', 'id', 'name', 'type_entity']
        assert fd_manager == ['id', 'manager_name', 'name', 'type_entity']

    def test_multi_table_poly(self, registry_multi_table_poly):
        registry = registry_multi_table_poly
        check_registry(registry.Employee)
        check_registry(registry.Engineer, engineer_name='An engineer')
        check_registry(registry.Manager, manager_name='An manager')

    def test_query_with_polymorphic(self, registry_multi_table_poly):
        registry = registry_multi_table_poly
        registry.Employee.insert(name='employee')
        registry.Engineer.insert(name='engineer', engineer_name='john')
        registry.Manager.insert(name='manager', manager_name='doe')
        assert registry.Employee.query().count() == 3
        for mapper in (registry.Engineer,
                       [registry.Engineer, registry.Manager], '*'):
            query = registry.Employee.query().with_polymorphic(mapper)
            query = query.filter(
                registry.Engineer.engineer_name == 'john')
            employee = query.one()
            assert isinstance(employee, registry.Engineer)

    def test_getFieldType(self, registry_multi_table_poly):
        registry = registry_multi_table_poly
        assert registry.Employee.getFieldType('id') == 'Integer'
        assert registry.Employee.getFieldType('name') == 'String'
        assert registry.Employee.getFieldType('type_entity') == 'String'
        assert registry.Engineer.getFieldType('id') == 'Integer'
        assert registry.Engineer.getFieldType('name') == 'String'
        assert registry.Engineer.getFieldType('type_entity') == 'String'
        assert registry.Engineer.getFieldType('engineer_name') == 'String'
        assert registry.Manager.getFieldType('id') == 'Integer'
        assert registry.Manager.getFieldType('name') == 'String'
        assert registry.Manager.getFieldType('type_entity') == 'String'
        assert registry.Manager.getFieldType('manager_name') == 'String'


def single_table_poly():
    @register(Model)
    class Employee:
        id = Integer(primary_key=True)
        type_entity = String(label="Entity type")
        name = String()
        engineer_name = String()
        manager_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Employee, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'employee',
                'polymorphic_on': cls.type_entity,
            })
            return mapper_args

    @register(Model, tablename=Model.Employee)
    class Engineer(Model.Employee):
        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Engineer, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'engineer',
            })
            return mapper_args

    @register(Model, tablename=Model.Employee)
    class Manager(Model.Employee):
        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Manager, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'manager',
            })
            return mapper_args


@pytest.fixture(scope="class")
def registry_single_table_poly(request, bloks_loaded):
    registry = init_registry(single_table_poly)
    request.addfinalizer(registry.close)
    return registry


class TestSingleTablePoly:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_single_table_poly):
        transaction = registry_single_table_poly.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_single_table_poly(self, registry_single_table_poly):
        registry = registry_single_table_poly
        check_registry(registry.Employee)
        check_registry(registry.Engineer, engineer_name='An engineer')
        check_registry(registry.Manager, manager_name='An manager')

    def test_get_primary_keys_on_single_table(self, registry_single_table_poly):
        registry = registry_single_table_poly
        employee_pks = registry.Employee.get_primary_keys()
        engineer_pks = registry.Engineer.get_primary_keys()
        assert employee_pks == engineer_pks


def multi_table_poly_mixins():
    @register(Mixin)
    class MyMixins:
        birthday = Date()

    @register(Model)
    class Employee:
        id = Integer(primary_key=True)
        name = String()
        type_entity = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Employee, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'employee',
                'polymorphic_on': cls.type_entity,
            })
            return mapper_args

    @register(Model)
    class Engineer(Model.Employee):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        engineer_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Engineer, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'engineer',
            })
            return mapper_args

    @register(Model)
    class Manager(Model.Employee, MyMixins):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        manager_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Manager, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'manager',
            })
            return mapper_args


@pytest.fixture(scope="class")
def registry_multi_table_poly_mixins(request, bloks_loaded):
    registry = init_registry(multi_table_poly_mixins)
    request.addfinalizer(registry.close)
    return registry


class TestMultiTablePolyMixins:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_multi_table_poly_mixins):
        transaction = registry_multi_table_poly_mixins.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_multi_table_poly_mixins(self, registry_multi_table_poly_mixins):
        registry = registry_multi_table_poly_mixins
        check_registry(registry.Employee)
        check_registry(registry.Engineer, engineer_name='An engineer')
        check_registry(registry.Manager, manager_name='An manager',
                       birthday=date.today())

    def test_getFieldType_with_mixin(self, registry_multi_table_poly_mixins):
        registry = registry_multi_table_poly_mixins
        assert registry.Employee.getFieldType('id') == 'Integer'
        assert registry.Employee.getFieldType('name') == 'String'
        assert registry.Employee.getFieldType('type_entity') == 'String'
        assert registry.Engineer.getFieldType('id') == 'Integer'
        assert registry.Engineer.getFieldType('name') == 'String'
        assert registry.Engineer.getFieldType('type_entity') == 'String'
        assert registry.Engineer.getFieldType('engineer_name') == 'String'
        assert registry.Manager.getFieldType('id') == 'Integer'
        assert registry.Manager.getFieldType('name') == 'String'
        assert registry.Manager.getFieldType('type_entity') == 'String'
        assert registry.Manager.getFieldType('manager_name') == 'String'
        assert registry.Manager.getFieldType('birthday') == 'Date'


def multi_table_foreign_key():
    @register(Model)
    class Room:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Employee:
        id = Integer(primary_key=True)
        name = String()
        type_entity = String()
        room = Many2One(model=Model.Room, one2many='employees')

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Employee, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'employee',
                'polymorphic_on': cls.type_entity,
            })
            return mapper_args

    @register(Model)
    class Engineer(Model.Employee):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        engineer_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Engineer, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'engineer',
            })
            return mapper_args


@pytest.fixture(scope="class")
def registry_multi_table_poly_fk(request, bloks_loaded):
    registry = init_registry(multi_table_foreign_key)
    request.addfinalizer(registry.close)
    return registry


class TestMultiTablePolyFk:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_multi_table_poly_fk):
        transaction = registry_multi_table_poly_fk.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_multi_table_foreign_key(self, registry_multi_table_poly_fk):
        registry = registry_multi_table_poly_fk
        room = registry.Room.insert()
        check_registry(registry.Employee, room=room)
        check_registry(registry.Engineer,
                       engineer_name='An engineer', room=room)

    def test_getFieldType_with_relationship(self, registry_multi_table_poly_fk):
        registry = registry_multi_table_poly_fk
        assert registry.Employee.getFieldType('id') == 'Integer'
        assert registry.Employee.getFieldType('name') == 'String'
        assert registry.Employee.getFieldType('type_entity') == 'String'
        assert registry.Employee.getFieldType('room') == 'Many2One'
        assert registry.Engineer.getFieldType('id') == 'Integer'
        assert registry.Engineer.getFieldType('name') == 'String'
        assert registry.Engineer.getFieldType('type_entity') == 'String'
        assert registry.Engineer.getFieldType('engineer_name') == 'String'
        assert registry.Engineer.getFieldType('room') == 'Many2One'


def multi_table_foreign_key_with_one_define_mapper_args():
    @register(Model)
    class Room:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Employee:
        id = Integer(primary_key=True)
        name = String()
        type_entity = String()
        room = Many2One(model=Model.Room, one2many='employees')

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Employee, cls).define_mapper_args()
            if Model.Employee.__registry_name__ == cls.__registry_name__:
                mapper_args.update({
                    'polymorphic_on': cls.type_entity,
                })

            mapper_args.update({
                'polymorphic_identity': cls.__registry_name__,
            })
            return mapper_args

    @register(Model)
    class Engineer(Model.Employee):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        engineer_name = String()


class TestPolymorphic:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_field_function_only_must_not_create_a_new_table(self):

        def single_table_poly_with_field_function():
            single_table_poly()

            @register(Model, tablename=Model.Employee)
            class Manager:

                full_name = Function(fget='get_full_name')

                def get_full_name(self):
                    return self.name + ' - ' + self.manager_name

        registry = self.init_registry(single_table_poly_with_field_function)
        check_registry(registry.Employee)
        check_registry(registry.Engineer, engineer_name='An engineer')
        t = check_registry(registry.Manager, manager_name='An manager')
        assert registry.Employee.__tablename__ == registry.Manager.__tablename__
        assert t.full_name == 'test - An manager'

    def test_field_function_must_be_available_in_subclass(self):

        def single_table_poly_with_field_function():
            single_table_poly()

            @register(Model)
            class Employee:

                full_name = Function(fget='get_full_name')

                def get_full_name(self):
                    sub_name = ''
                    if self.type_entity == 'manager':
                        sub_name = self.manager_name
                    elif self.type_entity == 'engineer':
                        sub_name = self.engineer_name

                    return self.name + ' - ' + sub_name

        registry = self.init_registry(single_table_poly_with_field_function)
        check_registry(registry.Employee)
        t1 = check_registry(registry.Engineer, engineer_name='An engineer')
        assert t1.full_name == 'test - An engineer'
        t2 = check_registry(registry.Manager, manager_name='An manager')
        assert t2.full_name == 'test - An manager'

    def test_multi_table_foreign_key2_with_one_define_mapper_args(self):
        registry = self.init_registry(
            multi_table_foreign_key_with_one_define_mapper_args)
        room = registry.Room.insert()
        check_registry(registry.Employee, room=room)
        check_registry(registry.Engineer, engineer_name='An engineer',
                       room=room)

    def test_query_with_polymorphic(self):
        registry = self.init_registry(multi_table_poly)
        registry.Employee.insert(name='employee')
        registry.Engineer.insert(name='engineer', engineer_name='john')
        registry.Manager.insert(name='manager', manager_name='doe')
        assert registry.Employee.query().count() == 3
        for mapper in (registry.Engineer,
                       [registry.Engineer, registry.Manager], '*'):
            query = registry.Employee.query().with_polymorphic(mapper)
            query = query.filter(
                registry.Engineer.engineer_name == 'john')
            employee = query.one()
            assert isinstance(employee, registry.Engineer)

    def test_getFieldType(self):
        registry = self.init_registry(multi_table_poly)
        assert registry.Employee.getFieldType('id') == 'Integer'
        assert registry.Employee.getFieldType('name') == 'String'
        assert registry.Employee.getFieldType('type_entity') == 'String'
        assert registry.Engineer.getFieldType('id') == 'Integer'
        assert registry.Engineer.getFieldType('name') == 'String'
        assert registry.Engineer.getFieldType('type_entity') == 'String'
        assert registry.Engineer.getFieldType('engineer_name') == 'String'
        assert registry.Manager.getFieldType('id') == 'Integer'
        assert registry.Manager.getFieldType('name') == 'String'
        assert registry.Manager.getFieldType('type_entity') == 'String'
        assert registry.Manager.getFieldType('manager_name') == 'String'

    def test_getFieldType_with_mixin(self):
        registry = self.init_registry(multi_table_poly_mixins)
        assert registry.Employee.getFieldType('id') == 'Integer'
        assert registry.Employee.getFieldType('name') == 'String'
        assert registry.Employee.getFieldType('type_entity') == 'String'
        assert registry.Engineer.getFieldType('id') == 'Integer'
        assert registry.Engineer.getFieldType('name') == 'String'
        assert registry.Engineer.getFieldType('type_entity') == 'String'
        assert registry.Engineer.getFieldType('engineer_name') == 'String'
        assert registry.Manager.getFieldType('id') == 'Integer'
        assert registry.Manager.getFieldType('name') == 'String'
        assert registry.Manager.getFieldType('type_entity') == 'String'
        assert registry.Manager.getFieldType('manager_name') == 'String'
        assert registry.Manager.getFieldType('birthday') == 'Date'

    def test_getFieldType_with_relationship(self):
        registry = self.init_registry(multi_table_foreign_key)
        assert registry.Employee.getFieldType('id') == 'Integer'
        assert registry.Employee.getFieldType('name') == 'String'
        assert registry.Employee.getFieldType('type_entity') == 'String'
        assert registry.Employee.getFieldType('room') == 'Many2One'
        assert registry.Engineer.getFieldType('id') == 'Integer'
        assert registry.Engineer.getFieldType('name') == 'String'
        assert registry.Engineer.getFieldType('type_entity') == 'String'
        assert registry.Engineer.getFieldType('engineer_name') == 'String'
        assert registry.Engineer.getFieldType('room') == 'Many2One'

    def test_field_description(self):
        registry = self.init_registry(multi_table_poly)
        fd_employee = list(registry.Employee.fields_description().keys())
        fd_employee.sort()
        fd_engineer = list(registry.Engineer.fields_description().keys())
        fd_engineer.sort()
        fd_manager = list(registry.Manager.fields_description().keys())
        fd_manager.sort()
        assert fd_employee == ['id', 'name', 'type_entity']
        assert fd_engineer == ['engineer_name', 'id', 'name', 'type_entity']
        assert fd_manager == ['id', 'manager_name', 'name', 'type_entity']

    def test_get_primary_keys_on_single_table(self):
        registry = self.init_registry(single_table_poly)
        employee_pks = registry.Employee.get_primary_keys()
        engineer_pks = registry.Engineer.get_primary_keys()
        assert employee_pks == engineer_pks

    def test_datetime_with_auto_update_on_single_table(self):

        def single_table_with_datetime_auto_update():
            single_table_poly()

            @register(Model)
            class Employee:
                update_at = DateTime(auto_update=True)

        registry = self.init_registry(single_table_with_datetime_auto_update)
        check_registry(registry.Employee)
        engineer = check_registry(
            registry.Engineer, engineer_name='An engineer')
        assert engineer.update_at is None
        manager = check_registry(
            registry.Manager, manager_name='An manager')
        assert manager.update_at is None
        engineer.engineer_name = 'Other'
        manager.manager_name = 'Other'
        registry.flush()
        assert engineer.update_at is not None
        assert manager.update_at is not None

    def test_datetime_with_auto_update_on_multi_table(self):

        def multi_table_with_datetime_auto_update():
            multi_table_poly()

            @register(Model)
            class Employee:
                update_at = DateTime(auto_update=True)

        registry = self.init_registry(multi_table_with_datetime_auto_update)
        check_registry(registry.Employee)
        engineer = check_registry(
            registry.Engineer, engineer_name='An engineer')
        assert engineer.update_at is None
        manager = check_registry(
            registry.Manager, manager_name='An manager')
        assert manager.update_at is None
        engineer.engineer_name = 'Other'
        manager.manager_name = 'Other'
        registry.flush()
        assert engineer.update_at is not None
        assert manager.update_at is not None
